;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((org-mode . ((org-adapt-indentation . nil)
              (org-edit-src-content-indentation . 0)))
 (nil . ((fill-column . 80))))
